/*
 * Copyright (c) 1998 Niklas Hallqvist, Tobias Weingartner
 * Copyright (c) 2002, 2003 Robert Watson, Michael Ranner
 * Copyright (c) 2004 Michael Ranner
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/fcntl.h>
#include <ufs/ufs/dinode.h>
#include <ufs/ffs/fs.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <err.h>
#if 0
#include <util.h>
#endif

#if __FreeBSD_version < 500037
#include "scan_ffs_freebsd5.h"
#endif

#if __FreeBSD_version >= 500037
#define SBSIZE SBLOCKSIZE
#endif
#define SBCOUNT 64		/* XXX - Should be configurable */

/* Flags to control ourselves... */
#define FLAG_VERBOSE		1
#define FLAG_SMART		2
#define FLAG_LABELS		4

static void usage(void);

static int
ufsscan(int fd, daddr_t beg, daddr_t end, int flags)
{
	static char lastmount[MAXMNTLEN];
	static u_int8_t buf[SBSIZE * SBCOUNT];
	struct fs *sb;
	daddr_t blk, lastblk;
	int n;
	int ufs;

	lastblk = -1;
	memset(lastmount, 0, MAXMNTLEN);

	for (blk = beg; blk <= ((end<0)?blk:end); blk += (SBCOUNT*SBSIZE/512)){
		memset(buf, 0, SBSIZE * SBCOUNT);
		if (lseek(fd, (off_t)blk * 512, SEEK_SET) < 0)
		    err(1, "lseek");
		if (read(fd, buf, SBSIZE * SBCOUNT) < 0)
			err(1, "read");

		for (n = 0; n < (SBSIZE * SBCOUNT); n += 512){
			ufs = 0;
			sb = (struct fs*)(&buf[n]);

			if (sb->fs_magic == FS_UFS1_MAGIC)
				ufs = 1;
#if __FreeBSD_version >= 500037
			if (sb->fs_magic == FS_UFS2_MAGIC)
				ufs = 2;
#endif

			if (ufs > 0) {
#if __FreeBSD_version >= 500037
				if (ufs == 1) sb->fs_size = sb->fs_old_size;
#endif
				if (flags & FLAG_VERBOSE)
#if __FreeBSD_version >= 500037
					printf("block %lld id %x,%x size %lld\n",
#else
					printf("block %d id %x,%x size %d\n",
#endif
					    blk + (n/512), sb->fs_id[0],
					    sb->fs_id[1], sb->fs_size);

				if (((blk+(n/512)) - lastblk) == (SBSIZE*ufs/512)) {
					if (flags & FLAG_LABELS ) {
#if __FreeBSD_version >= 500037
						printf("X: %lld %lld 4.2BSD %ld %ld %ld # %s\n",
#else
						printf("X: %d %d 4.2BSD %d %d %d # %s\n",
#endif
						    (daddr_t)((off_t)sb->fs_size *
						        sb->fs_fsize / 512),
		  				    blk+(n/(512))-(2*SBSIZE/512)-((ufs-1)*SBLOCK_UFS2/512),
						    sb->fs_fsize, sb->fs_bsize,
#if __FreeBSD_version < 500037
						    sb->fs_cpg, lastmount);
#else
						    sb->fs_old_cpg, lastmount);
#endif
					} else {
#if __FreeBSD_version >= 500037
						printf("ufs%d at %lld size %lld mount %s time %s\n",
#else
						printf("ufs%d at %d size %d mount %s time %s\n",
#endif
							ufs,
		  				    blk+(n/(512))-(2*SBSIZE/512)-((ufs-1)*SBLOCK_UFS2/512),
						    sb->fs_size,
						    lastmount, ctime((time_t *)&sb->fs_time));
					}

					if (flags & FLAG_SMART) {
						off_t size = (off_t)sb->fs_size *	
							sb->fs_fsize;

						if ((n + size) < (SBSIZE * SBCOUNT)) 
							n += size;
						else {
							blk += (size/512) - (SBCOUNT * SBCOUNT);
							break;
						}
					}
				}

				/* Update last potential FS SBs seen */
				lastblk = blk + (n/512);
				memcpy(lastmount, sb->fs_fsmnt, MAXMNTLEN);
			}
		}
	}
	return(0);
}


static void
usage(void)
{
	extern char *__progname;

	fprintf(stderr, "usage: %s [-lsv] [-b begin] [-e end] device\n",
	    __progname);
	exit(1);
}


int
main(int argc, char *argv[])
{
	char *name;
	int ch, fd, flags = 0;
	daddr_t beg = 0, end = -1;

	while ((ch = getopt(argc, argv, "lsvb:e:")) != -1)
		switch(ch) {
		case 'b':
			beg = atoi(optarg);
			break;
		case 'e':
			end = atoi(optarg);
			break;
		case 'v':
			flags |= FLAG_VERBOSE;
			break;
		case 's':
			flags |= FLAG_SMART;
			break;
		case 'l':
			flags |= FLAG_LABELS;
			break;
		default:
			usage();
			/* NOTREACHED */
	}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();

	if (argv[0][0] != '/') {
		if (asprintf(&name, "/dev/%s", argv[0]) == -1)
			err(1, "%s", argv[1]);
	} else
		name = argv[0];
	fd = open(name, O_RDONLY);
	if (fd < 0)
		err(1, "%s", argv[0]);
	if (name != argv[0])
		free(name);

	return (ufsscan(fd, beg, end, flags));
}
